// --- [ AT_Keyboard.c ] ------------------------------------------------------
//
// liest die Scan-/Break-Codes einer AT-Tastatur, 
// wandelt die Scan-Codes um in ASCII-Zeichen bzw. Steuer-Zeichen
// gibt diese via RS232 an den Host weiter
// Tabwidth = 3

// Anschluss Tastatur/Maus: clock an INT0, data an PD4 -> siehe kb.h
// notwendig ist ein Quarz fr den Betrieb der seriellen Schnittstelle
// Ausgabe mit 38400 Baud -> siehe rs232.h
// ----------------------------------------------------------------------------

#define EXTERN

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h> 
#include <util/delay.h>   
#include "kb.h"
#include "rs232.h"
#include "scancodes.h"

//#define SLEEP									// den Sleep-Modus aktivieren

volatile uint8_t control;						// enthlt den Status von CTRL, ALT, SHIFT etc.
volatile uint8_t rx_count;						// Zhler fr eingehende Bits, bentigt in INT0_ISR
volatile uint8_t rx_error = 0;				// hier werden Fehlercodes abgelegt

uint8_t kb_decode(unsigned char c);			// Wandelt SCAN-Code in ASCII-Code um

// ----------------------------------------------------------------------------
int main(void)
{
uint8_t i, j;

//TCCR0B = (1<<CS01 | 1<< CS00); 			// Prescale 64 (256 * 64 / 7372800 = 2.2ms)
TIMSK |= (1<<TOIE0);								// Timer0 Overflow enablen

//DDRB |= (1<<PB0 | 1<<PB1 | 1<<PB2 | 1<<PB3); // zum debuggen

ACSR = 1<<ACD;										// Analog-Comparator disablen
rs232_init();										// serielle Schnittstelle und tx_buffer initialisieren

put_c(13);
puts_p(PSTR("AT_Keyboard"));					// Hallo world
put_c(13);

PS2_init();              						// Initialize PS2-Schnittstelle
asm volatile("sei");             			// interrupt enable
 
while(1)
   {
	if (rx_error)									// wenn ein Fehler aufgetreten ist:
		{												// z.B. - Start- Paritts- oder Stop-Bit
		put_s(" Err: ");							// - timeout
		print_hexbyte(rx_error);				// - Fehler beim Senden
		rx_error = 0;								// Fehlerflag zurcksetzen
		//while (1);								// beenden - sofern gewnscht
		}
		
	if (PS2_is_byte_in_buffer())				// wenn Daten im Puffer sind
		{
		j = PS2_get_byte_from_buffer();		// j enthlt nun den SCAN-Code
//		print_hexbyte(j);							// Hexcode zum debuggen ausgeben


// --- [ SCAN-Code in ASCII-Code bersetzen ] ---------------------------------
		i = kb_decode(j);							// i enthlt nun den ASCII-Wert -					
														// oder 0, wenn eine Steuertaste gedrckt wurde
					
		if (i && (!(control & (1<<BREAK)))) // wenn BREAK nicht gesetzt ist
			{
			if ((i> 0x1D)) // && (i < 0x80)) // nur Code von 30 bis 127 drucken
				{
				put_c(i); 
				}
			else										// fr alles andere den Hex-Code
				{
				print_hexbyte(i);		
				}

//			print_hexbyte(control);				// Status von CTRL, ALT, SHIFT etc ausgeben
			}

		// wenn das aktuelle Zeichen kein EXT/BREAK-Code ist, dann EXT- und BREAK-Flag lschen
		if (!(j == 0xF0 || j == 0xE0)) control &= ~(1<< BREAK | 1<< EXT);
		}


#ifdef SLEEP
// bei der Ausgabe ber RS232 ist ein Quarz notwendig - und damit ist als SLEEP-Modus nur der
// IDLE-Modus einsetzbar
// kann man den RC-Oszillator einsetzen, dann gibt es greres Einsparpotential
// andererseits frit die Tastatur bei mir 170mA bei drei leuchtenden LEDs.
   else 												// wenn nichts zu tun ist und kein Sendeflag gesetzt, 
														// dann in SLEEP-Modus gehen
		{												// zuerst sicherstellen, dass nicht noch Bytes bertragen werden
		if (UCSRA & (1<<UDRE))					// hier die serielle Schnittstelle testen
			{											// wenn USART Data Register empty, dann ist bertragung beendet
			MCUCR |= (1<<SE);						// Sleep enable, wegen XTAL ist nur der IDLE-Modus mglich:
														// reduziert die Stromaufnahme von ca. 10mA auf 7mA.
			asm volatile("sleep"); 				// sleep, Aufwecken durch INT0-Interrupt
			}
		}
#endif
	}
}

// ----------------------------------------------------------------------------
// begrenzt den Zeitraum, in dem auf 11 eingehende Bits gewartet wird.
// Der Timer wird am Start einer Codesequenz zurckgesetzt.
// Wenn der Interrupt auslst, dann wird bit_count auf 11 gesetzt.
// Soll verhindern, dass die Bitzhlung aus dem Tritt kommt.
// Ein TimerOVF lt sich daran erkennen, dass TCCR0B auf 0 steht.

ISR(TIMER0_OVF_vect)
{
TCCR0B 	 = 0;									// Timer0 stoppen, 
													// anstelle eines Flags gleichzeitig Indikator fr TimerOVF
TCNT0  	 = 0;									// 
rx_count  = 11;								// Zhler zurcksetzen fr INT0
control  &= ~(1<<BREAK | 1<< EXT);		// Flag fr BREAK/EXT lschen
rx_error  = 0x10;								// Fehlercode fr TimerOVF speichern
}

// ----------------------------------------------------------------------------
// Wandelt den Tastataturcode in ein Zeichen um und bercksichtigt dabei,
// ob SHIFT (oder CAPSLOCK) gedrckt ist (-> unterschiedliche Zeichentabellen)
// Der Status von CTRL, ALT, SHIFT etc. steht in der Variablen "control".
// Wenn die bersetzung erfolgreich ist, wird der ASCII-Wert zurckgegeben - sonst "0"
uint8_t kb_decode(unsigned char c)
{
register uint8_t i, x = 0, t = 0;
static uint8_t led = 0;						// fr den Status der leds - als Demo

// zuerst auf die SteuerCodes testen, die nur Statusinformation ndern
switch (c)
	{
	case 0xE0:
	control |= (1<<EXT);						// Flag fr EXT setzen
	return(x);

	case 0xF0:
	control |= (1<<BREAK);					// Flag fr Break setzen
	return(x);

// die Aktionen der nachfolgenden Tasten sind abhngig davon, 
// ob Break gesetzt ist:
// beim Drcken wird aktiviert, beim Loslassen wieder deaktiviert

	case 0x12 :        						// Left SHIFT
	case 0x59 :        						// Right SHIFT
	if (!(control & (1<<BREAK)))			// bei make 
		{
		control |= (1<<SHIFT);				// SHIFT setzen
		}
	else											// bei break
		{
		control &= ~(1<<SHIFT);				// SHIFT lschen
		}		
	return(x);

	case 0x11:									// ALT left, ALT right + E0
	if (!(control & (1<<BREAK)))			// bei make
		{
		control |= (1<<ALT);					// ALT setzen	
		}
	else											// bei break
		{
		control &= ~(1<<ALT);				// ALT lschen
		}
	return(x);

	case 0x14:									// CTRL left, CTLR right + E0
	if (!(control & (1<<BREAK)))			// bei make
		{
		control |= (1<<CTRL);				// CTRL setzen
		}

	else											// bei break
		{
		control &= ~(1<<CTRL);				// CTRL lschen
		}
	return(x);
	}
	
if (c == 0x71)									// Prfen auf CTRL, ALT, DEL, hier auf DEL
	{												// bei DEL muss auch EXT gesetzt sein 
	if ( (control & (1<<CTRL | 1<<ALT | 1<<EXT )) == (1<<CTRL | 1<<ALT | 1<<EXT) )

		{
		control = 0; 							// CTRL und ALT manuell zurcksetzen,
		PS2_cmd(KB_RESET);					// denn durch den RESET fehlt der BREAK-Code !!
		return(0);
		}
	}
	
// wenn Break-Flag gesetzt ist, dann sind die nachfolgenden Codes ohne Bedeutung:
if (control & (1<<BREAK)) return(x);	// daher beenden

// hier nach Bedarf die Zeichen abfragen , fr die es keine Abhngigkeit von SHIFT ON/OFF gibt:
// wie RETURN, TAB, ESC, Backspace, Funktionstasten etc.
switch (c)
	{
	case 0x5a :									// RETURN

	led++;										// LED-Muster togglen als Demo
	led &= 7;			
		
	PS2_cmd(KB_LED); 							// das Kommando senden, tx_flag wird recycled
	PS2_cmd(led);								// den Parameter senden
	put_c(13);								// passend zum return ein CR senden
	
//	return(0x0D);								// wenn noch weitere Auswertung notwendig ist
	return(0);									// eine 0 tut nix weiteres mehr !
	
	case 0x29 :									// SPACE
	return(' ');
/*	
	case 0x0D :									// TAB
	return(0x09);
	case 0x76 :									// ESC
	return(0x1B);
	case 0x66 :									// BACKSPACE
	return(0x08);
	case 0x77 :									// NUMLOCK
	case 0x7E : 								// SCROLL
*/
	
	case 0x05 : 								// F1 druckt "F1"
	puts_p(PSTR(" F1 "));
	return(0);
	
/*	case 0x06 : 								// F2
	case 0x04 :									// F3
	case 0x0C :									// F4
	case 0x03 : 								// F5
	case 0x0B : 								// F6
	case 0x83 : 								// F7
	case 0x0A :									// F8
	case 0x01 ;									// F9
	case 0x09 : 								// F10
	case 0x78 : 								// F11
	case 0x07 : 								// F12
*/	
	}
// es wre noch zu testen, ob das EXT-Bit gesetzt ist
// z.B. um das DIvisionszeichen aus dem NUM-Block zu erkennen
// oder die Cursortasten von den ZIffern zu unterscheiden.
// oder zwischen DEL und ',' 
/*

if (control & (1<<EXT))
	{
	switch (c)
		{
		case 0x75 :								// Cursor up
		case 0x6B :								// Cursor left
		case 0x72 :								// Cursor down
		case 0x74 : 							// Cursor right
		case 0x4A :								// NUM /
		case 0x70 :								// Insert
		case 0x6C :								// HOME
		case 0x7D :								// PGUP
		case 0x71 : 							// Delete
		case 0x7A :								// PGUP
		}
	}
*/


// nun die Codes in der Tabelle nachschlagen, die von SHIFT abhngen
if (!(control & (1<<SHIFT)))				// wenn SHIFT nicht gesetzt ist
	{
	i = 0;
	do 
		{
		t = pgm_read_byte(&unshifted[i]);
		if (t == c) 
			{
			x = pgm_read_byte(&unshifted[++i]);
			break;
			}
		i += 2;				
		} while (t);
	}	

else                    					// wenn SHIFT gedrckt
	{ 
	i = 0;
	do 
		{
		t = pgm_read_byte(&shifted[i]);
		if (t == c) 
			{
			x = pgm_read_byte(&shifted[++i]);
			break;
			}
		i += 2;
		} while (t);			
	}
   
return(x);
} 


// --- [ eof ] ----------------------------------------------------------------
