// --- [ AT_Mouse.c ] ---------------------------------------------------------
//
// liest eine PS2-Maus aus, 
// gibt die Daten via RS232 an den Host weiter
// Tabwidth = 3

// Anschluss Tastatur/Maus: clock an INT0, data an PD4 -> siehe kb.h
// notwendig ist ein Quarz fr den Betrieb der seriellen Schnittstelle
// Ausgabe mit 38400 Baud -> siehe rs232.h
// ----------------------------------------------------------------------------

#define EXTERN
#define MSG											// Meldung bei der Initialisierung ausgeben

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h> 
#include <util/delay.h>
#include <avr/pgmspace.h>   
#include "mouse.h"
#include "rs232.h"

//#define SLEEP									// den Sleep-Modus aktivieren

volatile uint8_t rx_count;						// Zhler fr eingehende Bits, bentigt in INT0_ISR
volatile uint8_t rx_error = 0;				// hier werden Fehlercodes abgelegt
uint8_t  ms_ctrl;									// hier wird das 1.Byte des 3-Byte-Packets abgelegt
int8_t   ms_x_move;								// hier wird das 2.Byte des 3-Byte-Packets abgelegt
int8_t   ms_y_move;								// hier wird das 3.Byte des 3-Byte-Packets abgelegt

void print_ms_data(void);
void get_ms_device_type(void);
void delay_10ms(uint8_t);

// ----------------------------------------------------------------------------
int main(void)
{
uint8_t i = 0, j;

//TCCR0B = (1<<CS01 | 1<< CS00); // Prescale 64 (256 * 64 / 7372800 = 2.2ms)
TIMSK |= (1<<TOIE0);								// Timer0 Overflow enablen

//DDRB |= (1<<PB0 | 1<<PB1 | 1<<PB2 | 1<<PB3); // zum debuggen

ACSR = 1<<ACD;										// Analog-Comparator disablen
rs232_init();										// serielle Schnittstelle initialisieren

put_c(13);
puts_p(PSTR("PS2-Mouse "));					// Hallo world

delay_10ms(100); 									// abwarten, bis die Maus sich selbst initialisiert hat
														// alles, was die Maus sendet, wird hier noch nicht empfangen
PS2_init();              						// Initialize PS2-Schnittstelle
asm volatile("sei");             			// interrupt enable

#ifdef MSG											// nun ist der mc empfangsbereit
puts_p(PSTR("reset: "));
#endif

PS2_cmd(MS_RESET);								// jetzt sollte mit 0xFA, 0xAA, 0x00 geantwortet werden
														// wobei der Reset offensichtlich etwas lnger braucht ...
														// darum wird nachfolgend auf die Bytes gewartet
delay_10ms(100);									// und dann nachsehen, ob die 3 Byte die richtigen Werte haben

#ifdef MSG
if (PS2_is_byte_in_buffer()) 
	{
	i = PS2_get_byte_from_buffer();
	print_hexbyte(i);
	}
if (PS2_is_byte_in_buffer())
	{
	j = PS2_get_byte_from_buffer();
	print_hexbyte(j);
	i += j;
	}
if (PS2_is_byte_in_buffer())
	{
	j = PS2_get_byte_from_buffer();
	print_hexbyte(j);
	}	

if (i == (uint8_t)(0xFA + 0xAA)) puts_p(PSTR("ok")); else puts_p(PSTR("err"));
#endif

PS2_cmd(MS_SET_SAMPLE_RATE);					// Datenrate einstellen
PS2_cmd(10);										// auf den kleinsten Wert 10 Samples / Sekunde

#ifdef MSG
put_c(13);
puts_p(PSTR("set sample rate 10: ")); 
#endif

delay_10ms(10);
#ifdef mSG
while (PS2_is_byte_in_buffer())
	{
	j = PS2_get_byte_from_buffer();
	print_hexbyte(j);
	}
put_c(13);

puts_p(PSTR("set resolution 0: "));
#endif
PS2_cmd(MS_SET_RESOLUTION);
PS2_cmd(0x00);										// 1count/mm

delay_10ms(10);
#ifdef MSG
while (PS2_is_byte_in_buffer())
	{
	j = PS2_get_byte_from_buffer();
	print_hexbyte(j);
	}
put_c(13);
puts_p(PSTR("get mouse id: "));				// schon mal den Text ausgeben

PS2_cmd(MS_GET_DEVICE_ID);						// command senden

delay_10ms(10);									// die Antwort abwarten
while (PS2_is_byte_in_buffer())
	{
	i = PS2_get_byte_from_buffer();
	print_hexbyte(i);
	}

put_c(13);
puts_p(PSTR("set reporting on: "));
#endif
PS2_cmd(MS_ENABLE_DATA_REPORTING);			// Die Ausgabe der Daten aktivieren

delay_10ms(10);
#ifdef MSG
while (PS2_is_byte_in_buffer())
	{
	j = PS2_get_byte_from_buffer();
	print_hexbyte(j);
	}
put_c(13);

#else
PS2_clear_buffer();								// alle aufgelaufenen Antworten im Puffer lschen
#endif

puts_p(PSTR("ready"));
put_c(13);
i = 0;												// Zhler fr die Dreiergruppen, die nun immer folgen werden

while(1)
   {
	if (rx_error)									// wenn ein Fehler aufgetreten ist:
		{												// z.B. - Start- Paritts- oder Stop-Bit
		put_s(" Err: ");							// - timeout
		print_hexbyte(rx_error);				// - Fehler beim Senden
		rx_error = 0;								// Fehlerflag zurcksetzen
		//while (1);								// beenden - sofern gewnscht
		}
		
	if (PS2_is_byte_in_buffer())				// wenn Daten im Puffer sind
		{
		j = PS2_get_byte_from_buffer();		// j enthlt nun den SCAN-Code
//		print_hexbyte(j);							// hier werden die drei Maus-Bytes nacheinander ausgegeben

		if (i == 0) 								// und zur Auswertung an drei Variablen bergeben
			{
			ms_ctrl = j;
			i++;
			}

		else if (i == 1) 
			{
			ms_x_move = j;
			i++;
			}

		else if (i == 2)
			{
			ms_y_move = j;
			i = 0;
			print_ms_data();						// die Mausdaten im Klartext ausgeben
			put_c(13);
			}
		}

#ifdef SLEEP
   else 												// wenn nichts zu tun ist und kein Sendeflag gesetzt, 
														// dann in SLEEP-Modus gehen
		{												// zuerst sicherstellen, dass nicht noch Bytes bertragen werden
		if (UCSRA & (1<<UDRE))					// hier die serielle Schnittstelle testen
			{											// wenn USART Data Register empty, dann ist bertragung beendet
			MCUCR |= (1<<SE);						// Sleep enable, wegen XTAL ist nur der IDLE-Modus mglich:
														// reduziert die Stromaufnahme von ca. 10mA auf 7mA.
			asm volatile("sleep"); 				// sleep, Aufwecken durch INT0-Interrupt
			}
		}
#endif

	}
}


// ----------------------------------------------------------------------------
// Interpretiert die MAusdaten
// die Mausdaten selbst werden bereits in main ausgegeben
void print_ms_data(void)
{
static int16_t x_pos, y_pos;
char msg[4];

put_c(' ');
//if (ms_ctrl & (1<<Y_OVF)) puts_p(PSTR("Y_OVF "));	// Y-OVerflow-Bit
//if (ms_ctrl & (1<<X_OVF)) puts_p(PSTR("X_OVF "));	// X-OVerflow-Bit
//if (ms_ctrl & (1<<Y_SIGN)) puts_p(PSTR("y- "));		// Y-Negatives Vorzeichen
//if (ms_ctrl & (1<<X_SIGN)) puts_p(PSTR("x- "));		// X-Negatives Vorzeichen

//if (ms_y_move < -9) puts_p(PSTR(" down"));
//if (ms_y_move >  9) puts_p(PSTR(" up"));
//if (ms_x_move < -9) puts_p(PSTR(" left"));
//if (ms_x_move >  9) puts_p(PSTR(" right"));

#define X_LIMIT 500
#define Y_LIMIT 500

puts_p(PSTR("X: "));
x_pos += ms_x_move;												// rechts ist +, links -
if (x_pos < -X_LIMIT) x_pos = -X_LIMIT;
if (x_pos >  X_LIMIT) x_pos =  X_LIMIT;
put_s(itoa(x_pos,msg,10));										// X-Koordinaten ausgeben

puts_p(PSTR(" Y: "));											// oben ist +, unten -
y_pos += ms_y_move;
if (y_pos < -Y_LIMIT) y_pos = -Y_LIMIT;
if (y_pos >  Y_LIMIT) y_pos =  Y_LIMIT;
put_s(itoa(y_pos,msg,10));										// Y-Koordinaten ausgeben

if (ms_ctrl & (1<<BUTTON_M)) puts_p(PSTR(" M"));		// Mittlere Maustaste gedrckt
if (ms_ctrl & (1<<BUTTON_R)) puts_p(PSTR(" R"));		// Rechte   Maustaste gedrckt
if (ms_ctrl & (1<<BUTTON_L)) puts_p(PSTR(" L"));		// Linke    Maustaste gedrckt

}

// ----------------------------------------------------------------------------
// begrenzt den Zeitraum, in dem auf 11 eingehende Bits gewartet wird.
// Der Timer wird am Start einer Codesequenz zurckgesetzt.
// Wenn der Interrupt auslst, dann wird bit_count auf 11 gesetzt.
// Soll verhindern, dass die Bitzhlung aus dem Tritt kommt.
// Ein TimerOVF lt sich daran erkennen, dass TCCR0B auf 0 steht.

ISR(TIMER0_OVF_vect)
{
TCCR0B 	 = 0;										// Timer0 stoppen, 
														// anstelle eines Flags gleichzeitig Indikator fr TimerOVF
TCNT0  	 = 0;										// 
rx_count  = 11;									// Zhler zurcksetzen fr INT0
rx_error  = 0x10;									// Fehlercode fr TimerOVF speichern
}


// ---------------------------------------------------------------------------
void delay_10ms(uint8_t j)
{
uint8_t i;
for (i = 0; i < j; i++) _delay_ms(10); 	//maximal 262ms / 1 MHz
}

// --- [ eof ] ----------------------------------------------------------------
